@@ -37,34 +37,48 @@ DENY = 0
 ACCEPT = 1
 OPEN = 2
 
+def basename(filename):
+    #we can't use os.path.basename,
+    #because the remote end may have sent us a filename
+    #which is using a different pathsep
+    for sep in ("\\", "/", os.sep):
+        i = filename.rfind(sep) + 1
+        filename = filename[i:]
+    if not WIN32:
+        return filename
+    #many characters aren't allowed at all on win32:
+    tmp = ""
+    for char in filename:
+        if ord(char)<32 or char in ("<", ">", ":", "\"", "|", "?", "*"):
+            char = "_"
+        tmp += char
+    return tmp
 
 def safe_open_download_file(basefilename, mimetype):
     from xpra.platform.paths import get_download_dir
-    #make sure we use a filename that does not exist already:
     dd = os.path.expanduser(get_download_dir())
-    base = os.path.basename(basefilename)
-    wanted_filename = os.path.abspath(os.path.join(dd, base))
+    filename = os.path.abspath(os.path.join(dd, basename(basefilename)))
     ext = MIMETYPE_EXTS.get(mimetype)
-    if ext:
+    if ext and not filename.endswith("."+ext):
         #on some platforms (win32),
         #we want to force an extension
         #so that the file manager can display them properly when you double-click on them
-        if not wanted_filename.endswith("."+ext):
-            wanted_filename += "."+ext
-    filename = wanted_filename
+        filename += "."+ext
+    #make sure we use a filename that does not exist already:
+    root, ext = os.path.splitext(filename)
     base = 0
     while os.path.exists(filename):
         filelog("cannot save file as %s: file already exists", filename)
-        root, ext = os.path.splitext(wanted_filename)
         base += 1
         filename = root+("-%s" % base)+ext
+    filelog("safe_open_download_file(%s, %s) will use '%s'", basefilename, mimetype, filename)
     flags = os.O_CREAT | os.O_RDWR | os.O_EXCL
     try:
         flags |= os.O_BINARY                #@UndefinedVariable (win32 only)
-    except:
+    except AttributeError:
         pass
     fd = os.open(filename, flags)
-    filelog("using filename '%s'", filename)
+    filelog("using filename '%s', file descriptor=%s", filename, fd)
     return filename, fd
 
 
@@ -78,7 +92,8 @@ class FileTransferAttributes(object):
         self.init_attributes(opts.file_transfer, opts.file_size_limit,
                              opts.printing, opts.open_files, opts.open_url, opts.open_command, can_ask)
 
-    def init_attributes(self, file_transfer="yes", file_size_limit=10, printing="yes", open_files="no", open_url="yes", open_command=None, can_ask=True):
+    def init_attributes(self, file_transfer="no", file_size_limit=10, printing="no",
+                        open_files="no", open_url="no", open_command=None, can_ask=True):
         filelog("file transfer: init_attributes%s",
                 (file_transfer, file_size_limit, printing, open_files, open_url, open_command, can_ask))
         def pbool(name, v):
@@ -179,8 +194,8 @@ class FileTransferHandler(FileTransferAttributes):
         self.receive_chunks_in_progress = {}
         for x in tuple(self.file_descriptors):
             try:
-                x.close()
-            except:
+                os.close(x)
+            except (OSError, IOError):
                 pass
         self.file_descriptors = set()
         self.init_attributes()
@@ -212,18 +227,18 @@ class FileTransferHandler(FileTransferAttributes):
     def get_info(self):
         info = FileTransferAttributes.get_info(self)
         info["remote"] = {
-                          "file-transfer"   : self.remote_file_transfer,
-                          "file-transfer-ask" : self.remote_file_transfer_ask,
-                          "file-size-limit" : self.remote_file_size_limit,
-                          "file-chunks"     : self.remote_file_chunks,
-                          "open-files"      : self.remote_open_files,
-                          "open-files-ask"  : self.remote_open_files_ask,
-                          "open-url"        : self.remote_open_url,
-                          "open-url-ask"    : self.remote_open_url_ask,
-                          "printing"        : self.remote_printing,
-                          "printing-ask"    : self.remote_printing_ask,
-                          "file-ask-timeout" : self.remote_file_ask_timeout,
-                          }
+            "file-transfer"     : self.remote_file_transfer,
+            "file-transfer-ask" : self.remote_file_transfer_ask,
+            "file-size-limit"   : self.remote_file_size_limit,
+            "file-chunks"       : self.remote_file_chunks,
+            "open-files"        : self.remote_open_files,
+            "open-files-ask"    : self.remote_open_files_ask,
+            "open-url"          : self.remote_open_url,
+            "open-url-ask"      : self.remote_open_url_ask,
+            "printing"          : self.remote_printing,
+            "printing-ask"      : self.remote_printing_ask,
+            "file-ask-timeout"  : self.remote_file_ask_timeout,
+            }
         return info
 
     def check_digest(self, filename, digest, expected_digest, algo="sha1"):
@@ -275,7 +290,7 @@ class FileTransferHandler(FileTransferAttributes):
             del self.receive_chunks_in_progress[chunk_id]
             try:
                 os.close(fd)
-            except:
+            except (OSError, IOError):
                 pass
             return
         self.send("ack-file-chunk", chunk_id, True, "", chunk)
@@ -299,6 +314,7 @@ class FileTransferHandler(FileTransferAttributes):
             self.check_digest(filename, digest.hexdigest(), expected_digest)
         start_time = chunk_state[0]
         elapsed = monotonic_time()-start_time
+        mimetype = bytestostr(mimetype)
         filelog("%i bytes received in %i chunks, took %ims", filesize, chunk, elapsed*1000)
         t = start_thread(self.do_process_downloaded_file, "process-download", daemon=False,
                          args=(filename, mimetype, printit, openit, filesize, options))
@@ -345,18 +361,27 @@ class FileTransferHandler(FileTransferAttributes):
             l.error("Error: file '%s' is too large:", basefilename)
             l.error(" %iMB, the file size limit is %iMB", filesize//1024//1024, self.file_size_limit)
             return
+        chunk_id = options.strget("file-chunk-id")
         #basefilename should be utf8:
         try:
             base = basefilename.decode("utf8")
         except:
             base = bytestostr(basefilename)
-        filename, fd = safe_open_download_file(base, mimetype)
+        try:
+            filename, fd = safe_open_download_file(base, mimetype)
+        except OSError as e:
+            filelog("cannot save file %s / %s", base, mimetype, exc_info=True)
+            filelog.error("Error: failed to save downloaded file")
+            filelog.error(" %s", e)
+            if chunk_id:
+                self.send("ack-file-chunk", chunk_id, False, "failed to create file: %s" % e, 0)
+            return
         self.file_descriptors.add(fd)
-        chunk_id = options.strget("file-chunk-id")
         if chunk_id:
             chunk_id = strtobytes(chunk_id)
-            if len(self.receive_chunks_in_progress)>=MAX_CONCURRENT_FILES:
-                self.send("ack-file-chunk", chunk_id, False, "too many file transfers in progress", 0)
+            l = len(self.receive_chunks_in_progress)
+            if l>=MAX_CONCURRENT_FILES:
+                self.send("ack-file-chunk", chunk_id, False, "too many file transfers in progress: %i" % l, 0)
                 os.close(fd)
                 return
             digest = hashlib.sha1()
@@ -424,7 +449,7 @@ class FileTransferHandler(FileTransferAttributes):
             if DELETE_PRINTER_FILE:
                 try:
                     os.unlink(filename)
-                except:
+                except (OSError, IOError):
                     printlog("failed to delete print job file '%s'", filename)
         if not printer:
             printlog.error("Error: the printer name is missing")
@@ -437,7 +462,7 @@ class FileTransferHandler(FileTransferAttributes):
             delfile()
             return
         try:
-            job_options = options.dictget("options")
+            job_options = options.dictget("options", {})
             job_options["copies"] = copies
             job = print_files(printer, [filename], title, job_options)
         except Exception as e:
@@ -697,17 +722,16 @@ class FileTransferHandler(FileTransferAttributes):
             filelog.error("Error: unknown datatype '%s'", dtype)
 
     def send_data_ask_timeout(self, send_id):
-        v = self.pending_send_data_timers.get(send_id)
-        if not v:
-            filelog.warn("Warning: send timeout, id '%s' not found!", send_id)
-            return False
+        v = self.pending_send_data.pop(send_id, None)
         try:
-            del self.pending_send_data[send_id]
             del self.pending_send_data_timers[send_id]
         except KeyError:
             pass
-        filename = v[0]
-        printit = v[4]
+        if not v:
+            filelog.warn("Warning: send timeout, id '%s' not found!", send_id)
+            return False
+        filename = v[1]
+        printit = v[5]
         filelog.warn("Warning: failed to %s file '%s',", ["send", "print"][printit], filename)
         filelog.warn(" the send approval request timed out")
         return False
@@ -759,7 +783,10 @@ class FileTransferHandler(FileTransferAttributes):
             chunk_state[-2] = 0         #timer has fired
             if chunk_state[-1]==chunk_no:
                 filelog.error("Error: chunked file transfer timed out on chunk %i", chunk_no)
-                del self.send_chunks_in_progress[chunk_id]
+                try:
+                    del self.send_chunks_in_progress[chunk_id]
+                except KeyError:
+                    pass
 
     def _process_ack_file_chunk(self, packet):
         #the other end received our send-file or send-file-chunk,
@@ -798,3 +825,9 @@ class FileTransferHandler(FileTransferAttributes):
         timer = self.timeout_add(CHUNK_TIMEOUT, self._check_chunk_sending, chunk_id, chunk)
         self.send_chunks_in_progress[chunk_id] = [start_time, data, chunk_size, timer, chunk]
         self.send("send-file-chunk", chunk_id, chunk, cdata, bool(data))
+
+    def send(self, *parts):
+        raise NotImplementedError()
+
+    def compressed_wrapper(self, datatype, data, level=5):
+        raise NotImplementedError()
