@@ -125,21 +125,26 @@ function XpraWindow(client, canvas_state, wid, x, y, w, h, metadata, override_re
 	}
 	else if((this.windowtype == "") || (this.windowtype == "NORMAL") || (this.windowtype == "DIALOG") || (this.windowtype == "UTILITY")) {
 		this.resizable = true;
+		jQuery(this.div).addClass("border");
 		// add a title bar to this window if we need to
 		// create header
 		jQuery(this.div).prepend('<div id="head' + String(wid) + '" class="windowhead"> '+
 				'<span class="windowicon"><img id="windowicon' + String(wid) + '" /></span> '+
 				'<span class="windowtitle" id="title' + String(wid) + '">' + this.title + '</span> '+
 				'<span class="windowbuttons"> '+
+				'<span id="minimize' + String(wid) + '"><img src="icons/minimize.png" /></span> '+
 				'<span id="maximize' + String(wid) + '"><img src="icons/maximize.png" /></span> '+
 				'<span id="close' + String(wid) + '"><img src="icons/close.png" /></span> '+
 				'</span></div>');
 		// make draggable
 		jQuery(this.div).draggable({ cancel: "canvas" });
 		jQuery(this.div).on("dragstart",function(ev,ui){
-			set_focus_cb(me);
+			client.do_window_mouse_click(ev, me, false);
+			client.mouse_grabbed = true;
+			me.set_focus_cb(me);
 		});
 		jQuery(this.div).on("dragstop",function(ev,ui){
+			client.mouse_grabbed = false;
 			me.handle_moved(ui);
 		});
 		// attach resize handles
@@ -147,30 +152,58 @@ function XpraWindow(client, canvas_state, wid, x, y, w, h, metadata, override_re
 		//jQuery(this.div).on("resize",jQuery.debounce(50, function(ev,ui) {
 		//  	me.handle_resized(ui);
 		//}));
+		jQuery(this.div).on("resizestart",function(ev,ui){
+			client.do_window_mouse_click(ev, me, false);
+			client.mouse_grabbed = true;
+		});
 		jQuery(this.div).on("resizestop",function(ev,ui){
 		  	me.handle_resized(ui);
-		  	set_focus_cb(me);
+		  	me.set_focus_cb(me);
+			client.mouse_grabbed = false;
+			//workaround for the window going blank,
+			//just force a refresh:
+			setTimeout(function() {
+				me.client.request_refresh(me.wid);
+			}, 200);
 		});
 		this.d_header = '#head' + String(wid);
 		this.d_closebtn = '#close' + String(wid);
 		this.d_maximizebtn = '#maximize' + String(wid);
+		this.d_minimizebtn = '#minimize' + String(wid);
 		if (this.resizable) {
+			jQuery(this.d_header).dblclick(function() {
+				me.toggle_maximized();
+			});
 			jQuery(this.d_closebtn).click(function() {
 				window_closed_cb(me);
 			});
 			jQuery(this.d_maximizebtn).click(function() {
 				me.toggle_maximized();
 			});
+			jQuery(this.d_minimizebtn).click(function() {
+				me.toggle_minimized();
+			});
 		}
 		else {
 			jQuery(this.d_closebtn).hide();
 			jQuery(this.d_maximizebtn).hide();
+			jQuery('#windowlistitemmax' + String(wid)).hide();
+			jQuery(this.d_minimizebtn).hide();
 		}
 		// adjust top offset
 		this.topoffset = this.topoffset + parseInt(jQuery(this.d_header).css('height'), 10);
+		// stop propagation if we're over the window:
+		jQuery(this.div).mousedown(function (e) {
+			e.stopPropagation();
+		});
+		//bug 2418: if we stop 'mouseup' propagation,
+		//jQuery can't ungrab the window with Firefox
+		//jQuery(this.div).mouseup(function (e) {
+		//	e.stopPropagation();
+		//});
 		// assign some interesting callbacks
 		jQuery(this.d_header).click(function() {
-			set_focus_cb(me);
+			me.set_focus_cb(me);
 		});
 	}
 
@@ -248,9 +281,10 @@ XpraWindow.prototype._init_2d_canvas = function() {
 
 XpraWindow.prototype.swap_buffers = function() {
 	//the up to date canvas is what we'll draw on screen:
+	this.debug("draw", "swap_buffers");
 	this.draw_canvas = this.offscreen_canvas;
 	this._init_2d_canvas();
-	this.offscreen_canvas_ctx.drawImage(this.draw_canvas, 0, 0);	
+	this.offscreen_canvas_ctx.drawImage(this.draw_canvas, 0, 0);
 }
 
 XpraWindow.prototype.set_spinner = function(state) {
@@ -436,7 +470,10 @@ XpraWindow.prototype.update_metadata = function(metadata, safe) {
 XpraWindow.prototype.set_metadata_safe = function(metadata) {
 	if ("title" in metadata) {
 		this.title = metadata["title"];
-		jQuery('#title' + this.wid).html(decodeURIComponent(escape(this.title)));
+		var decodedTitle = decodeURIComponent(escape(this.title));
+		jQuery('#title' + this.wid).html(decodedTitle);
+        var trimmedTitle = Utilities.trimString(decodedTitle,30);
+		jQuery('#windowlistitemtitle'+this.wid).text(trimmedTitle);
 	}
 	if ("has-alpha" in metadata) {
 		this.has_alpha = metadata["has-alpha"];
@@ -535,6 +572,7 @@ XpraWindow.prototype.apply_size_constraints = function() {
 	}
 	if(minw>0 && minw==maxw && minh>0 && minh==maxh) {
 		jQuery(this.d_maximizebtn).hide();
+		jQuery('#windowlistitemmax' + String(this.wid)).hide();
 		jQuery(this.div).resizable('disable');
 	} else {
 		jQuery(this.d_maximizebtn).show();
@@ -603,6 +641,10 @@ XpraWindow.prototype.restore_geometry = function() {
  * Maximize / unmaximizes the window.
  */
 XpraWindow.prototype.set_maximized = function(maximized) {
+	if(jQuery(this.div).is(":hidden")){
+		jQuery(this.div).show();
+	}
+
 	if (this.maximized==maximized) {
 		return;
 	}
@@ -621,6 +663,21 @@ XpraWindow.prototype.toggle_maximized = function() {
 	this.set_maximized(!this.maximized);
 };
 
+/**
+ * Minimizes / unminimizes the window.
+ */
+XpraWindow.prototype.set_minimized = function(minimized) {
+	jQuery(this.div).toggle(200);
+};
+
+
+/**
+ * Toggle minimized state
+ */
+XpraWindow.prototype.toggle_minimized = function() {
+	this.set_minimized(!this.minimized);
+};
+
 /**
  * Fullscreen / unfullscreen the window.
  */
@@ -766,13 +823,13 @@ XpraWindow.prototype.screen_resized = function() {
 	this.ensure_visible();
 };
 
-XpraWindow.prototype.recenter = function() {
+XpraWindow.prototype.recenter = function(force_update_geometry) {
 	var x = this.x,
 		y = this.y;
 	this.debug("geometry", "recenter() x=", x, ", y=", y, ", desktop size: ", this.client.desktop_width, this.client.desktop_height);
 	x = Math.round((this.client.desktop_width-this.w)/2);
 	y = Math.round((this.client.desktop_height-this.h)/2);
-	if (this.x!=x || this.y!=y) {
+	if (this.x!=x || this.y!=y || force_update_geometry) {
 		this.debug("geometry", "window re-centered to:", x, y);
 		this.x = x;
 		this.y = y;
@@ -834,7 +891,9 @@ XpraWindow.prototype.match_screen_size = function() {
 		}
 		this.log("best screen size:", neww, newh);
 	}
-	this.recenter();
+	this.w = neww;
+	this.h = newh;
+	this.recenter(true);
 };
 
 
@@ -928,9 +987,17 @@ XpraWindow.prototype.update_icon = function(width, height, encoding, img_data) {
 	if (encoding=="png") {
 		//move title to the right:
 		$("#title"+ String(this.wid)).css('left', 32);
+		if (typeof img_data === 'string') {
+			var uint = new Uint8Array(img_data.length);
+			for(var i=0;i<img_data.length;++i) {
+				uint[i] = img_data.charCodeAt(i);
+			}
+			img_data = uint;
+		}
 		src = "data:image/"+encoding+";base64," + Utilities.ArrayBufferToBase64(img_data);
 	}
 	jQuery('#windowicon' + String(this.wid)).attr('src', src);
+	jQuery('#windowlistitemicon' + String(this.wid)).attr('src', src);
 	return src;
 };
 
